import os
import sys

from jinjaform import log


args = sys.argv[1:]
cwd = os.getcwd()
env = os.environ.copy()

for name in ('JINJAFORM_PROJECT_ROOT', 'JINJAFORM_TERRAFORM_BIN'):
    if not env.get(name):
        log.bad('{} environment variable missing', name)
        sys.exit(1)

project_root = os.environ['JINJAFORM_PROJECT_ROOT']
jinjaform_root = os.path.join(project_root, '.jinjaform')
terraform_bin = os.environ['JINJAFORM_TERRAFORM_BIN']
workspace_dir = os.path.join(cwd, '.jinjaform')
terraform_dir = os.path.join(workspace_dir, '.terraform')

aws_provider = {}
s3_backend = {}
sessions = {}

env['JINJAFORM_PROJECT_ROOT'] = project_root
env['JINJAFORM_WORKSPACE'] = workspace_dir



import os
import sys
import subprocess

from jinjaform import log
from jinjaform.config import args


def abort():
    if os.environ.get('GIT_CHECK') == '0':
        pass
    else:
        log.bad('git: refusing to continue')
        log.ok('hint: set GIT_CHECK=0 to bypass git checks')
        sys.exit(1)


def check_branch(desired):

    if 'apply' not in args:
        return

    branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).rstrip().decode('utf-8')
    if branch != desired:
        log.bad('git: working in branch {}', branch)
        log.bad('git: should be in branch {}', desired)
        abort()


def check_clean():

    if 'apply' not in args:
        return

    diff = subprocess.check_output(['git', 'status', '--porcelain']).rstrip().decode('utf-8')
    if diff:
        log.bad('git: working directory not clean')
        for line in diff.splitlines():
            log.bad('git: {}', line)
        abort()


def check_remote():

    if 'apply' not in args:
        return

    log.ok('git: checking remote')
    subprocess.check_output(['git', 'remote', 'update'])
    local = subprocess.check_output(['git', 'rev-parse', 'HEAD']).rstrip().decode('utf-8')

    try:
        remote = subprocess.check_output(
            ['git', 'rev-parse', '@{upstream}'],
            stderr=subprocess.STDOUT,
        ).rstrip().decode('utf-8')
    except subprocess.CalledProcessError as error:
        # There may not be any upstream configured for the branch.
        output = error.output.rstrip().decode('utf-8')
        log.bad('git: {}', output)
        abort()
    else:
        if local != remote:
            status = subprocess.check_output(['git', 'status']).rstrip().decode('utf-8')
            for line in status.splitlines():
                if 'branch' in line and ('ahead' in line or 'behind' in line):
                    log.bad('git: {}', line.lower())
                    break
            else:
                log.bad('git: out of date')
            abort()



import colorama


def init(_cache=[]):
    if not _cache:
        colorama.init()
        _cache.append(True)


def bad(message, *args, **kwargs):
    init()
    if args or kwargs:
        message = message.format(*args, **kwargs)
    print(colorama.Fore.RED + '[jinjaform] ' + message + colorama.Style.RESET_ALL)


def ok(message, *args, **kwargs):
    init()
    if args or kwargs:
        message = message.format(*args, **kwargs)
    print(colorama.Fore.CYAN + '[jinjaform] ' + message + colorama.Style.RESET_ALL)



import boto3
import botocore
import os
import sys

from jinjaform import log
from jinjaform.config import aws_provider, env, s3_backend, sessions


def backend_setup():

    region = s3_backend.get('region')
    if not region:
        return

    bucket = s3_backend.get('bucket')
    if bucket:

        log.ok('backend: s3://{} in {}', bucket, region)

        if sessions:
            s3_client = next(iter(sessions.values())).client('s3', region_name=region)
        else:
            s3_client = boto3.client('s3', region_name=region)

        try:
            response = s3_client.get_bucket_versioning(
                Bucket=bucket
            )
        except s3_client.exceptions.NoSuchBucket:
            bucket_exists = False
            bucket_versioning = False
        else:
            bucket_exists = True
            bucket_versioning = response['Status'] == 'Enabled'

        if not bucket_exists:

            question = 'create s3://{} in {} [yes/no]: '.format(bucket, region)
            answer = ''
            while answer not in ('yes', 'no'):
                answer = input(question).lower()
            if answer == 'no':
                log.ok('backend: bucket not created')
                sys.exit(1)

            log.ok('backend: creating bucket')
            s3_client.create_bucket(
                ACL='private',
                Bucket=bucket,
                CreateBucketConfiguration={
                    'LocationConstraint': region,
                },
            )
            s3_client.get_waiter('bucket_exists').wait(Bucket=bucket)

        if not bucket_versioning:
            log.ok('backend: enabling versioning')
            s3_client.put_bucket_versioning(
                Bucket=bucket,
                VersioningConfiguration={
                    'Status': 'Enabled',
                },
            )

    dynamodb_table = s3_backend.get('dynamodb_table')
    if dynamodb_table:

        log.ok('backend: dynamodb://{} in {}', dynamodb_table, region)

        if sessions:
            dynamodb_client = next(iter(sessions.values())).client('dynamodb', region_name=region)
        else:
            dynamodb_client = boto3.client('dynamodb', region_name=region)

        try:
            dynamodb_client.describe_table(
                TableName=dynamodb_table,
            )
        except dynamodb_client.exceptions.ResourceNotFoundException:

            question = 'create dynamodb://{} in {} [yes/no]: '.format(dynamodb_table, region)
            answer = ''
            while answer not in ('yes', 'no'):
                answer = input(question).lower()
            if answer == 'no':
                log.ok('backend: table not created')
                sys.exit(1)

            log.ok('creating table')
            dynamodb_client.create_table(
                TableName=dynamodb_table,
                AttributeDefinitions=[
                    {
                        'AttributeName': 'LockID',
                        'AttributeType': 'S',
                    }
                ],
                KeySchema=[
                    {
                        'AttributeName': 'LockID',
                        'KeyType': 'HASH',
                    },
                ],
                ProvisionedThroughput={
                    'ReadCapacityUnits': 1,
                    'WriteCapacityUnits': 1,
                },
            )


def credentials_setup():
    """
    Sets up AWS credentials using Terraform AWS provider blocks.

    """

    profile = aws_provider.get('profile')
    if not profile:
        return

    log.ok('aws-profile: {}', profile)

    botocore_session = botocore.session.Session(profile=profile)
    cli_cache_path = os.path.join(os.path.expanduser('~'), '.aws/cli/cache')
    cli_cache = botocore.credentials.JSONFileCache(cli_cache_path)
    botocore_session.get_component('credential_provider').get_provider('assume-role').cache = cli_cache
    session = boto3.Session(botocore_session=botocore_session)

    sessions[profile] = session

    try:
        creds = session.get_credentials().get_frozen_credentials()
    except KeyboardInterrupt:
        print()
        log.bad('aborted')
        sys.exit(1)

    env['AWS_PROFILE'] = profile
    env['AWS_DEFAULT_PROFILE'] = profile
    env['AWS_ACCESS_KEY_ID'] = creds.access_key
    env['AWS_SECRET_ACCESS_KEY'] = creds.secret_key
    env['AWS_SESSION_TOKEN'] = creds.token



import os
import subprocess
import sys

from jinjaform import aws, git, log, rc, terraform, workspace
from jinjaform.config import args, cwd, env, project_root, terraform_bin, workspace_dir


commands_bypassed = (
    'fmt',
    'help',
    '-help',
    'version',
    '-version',
)

commands_forbidden = (
    'push',
)

commands_using_backend = (
    'apply',
    'console',
    'debug'
    'destroy',
    'force-unlock',
    'graph',
    'import',
    'init',
    'output',
    'plan',
    'providers',
    'refresh',
    'show',
    'state',
    'taint',
    'untaint',
)

workspace_required = False

if args:

    if not set(commands_bypassed).intersection(args):

        if cwd != project_root and (cwd + '/').startswith(project_root + '/'):

            for command in commands_forbidden:
                if command in args:
                    log.bad('{} not allowed', command)
                    sys.exit(1)

            workspace_required = True

        else:

            log.bad('not in deployment target directory, aborting')
            sys.exit(1)

if workspace_required:

    for rc_cmd, rc_arg in rc.read():

        if rc_cmd == 'GIT_CHECK_BRANCH':

            git.check_branch(desired=rc_arg)

        elif rc_cmd == 'GIT_CHECK_CLEAN':

            git.check_clean()

        elif rc_cmd == 'GIT_CHECK_REMOTE':

            git.check_remote()

        elif rc_cmd == 'RUN':

            log.ok('run: {}'.format(rc_arg))
            returncode = subprocess.call(rc_arg, env=env, shell=True)
            if returncode != 0:
                sys.exit(returncode)

        elif rc_cmd == 'TERRAFORM_RUN':

            log.ok('run: terraform')
            os.chdir(workspace_dir)
            returncode = terraform.execute(terraform_bin, args, env)
            if returncode != 0:
                sys.exit(returncode)

        elif rc_cmd == 'WORKSPACE_CREATE':

            workspace.clean()
            workspace.create()

            if set(commands_using_backend).intersection(args):
                aws.credentials_setup()
                if 'init' in args:
                    aws.backend_setup()

        else:

            log.bad('configuration: {} is not a valid command', rc_cmd)
            sys.exit(1)

else:

    log.ok('run: terraform')
    returncode = terraform.execute(terraform_bin, args, env)
    if returncode != 0:
        sys.exit(returncode)



import hcl
import os
import shutil
import sys

from collections import defaultdict
from contextlib import suppress

from itertools import chain

from jinja2 import StrictUndefined, Template
from jinja2.exceptions import UndefinedError

from jinjaform import log, config
from jinjaform.config import cwd, env, jinjaform_root, project_root, terraform_dir, workspace_dir

from threading import current_thread, Event, Lock, Thread


class VarStore(object):

    def __init__(self, threads):

        self._threads = threads

        self._lock = Lock()
        self._waiters = defaultdict(list)

        self._defined = set()
        self._defaults = dict()
        self._values = dict()

        self._unresolved = dict()

    def __getitem__(self, key):
        """
        Returns a variable value. If the variable has not been defined
        then this will block and wait for it.

        """

        # Wait if the variable has not been defined yet.
        self._wait_for_variable(key)

        # Return the value if the variable has been defined
        # with a specific value or a default value.
        if key in self._defined:
            if key in self._values:
                return self._values[key]
            if key in self._defaults:
                return self._defaults[key]

        # The variable was not defined, or it was defined without a value,
        # or it was defined but there was a deadlock when resolving it.
        self._unresolved[current_thread()] = key
        raise KeyError(key)

    def _check_deadlock(self):
        """
        Checks if all threads are waiting for a variable to be defined.
        This means that the variable is not defined, or there is a circular
        dependency between files that define and use variables.

        If this is the case, then this function will unblock the threads,
        allowing them to continue executing without their variable. They
        would then have an error because the variable isn't defined.

        """

        waiters = list(chain.from_iterable(self._waiters.values()))
        all_threads_blocked = len(waiters) == len(self._threads)
        if all_threads_blocked:
            for event in waiters:
                event.set()

    def _define_variable(self, name, default):
        """
        Adds a variable definition and unblocks any waiting threads.

        """

        with self._lock:
            self._defined.add(name)
            if default is not None:
                self._defaults[name] = default
                for event in self._waiters.pop(name, []):
                    event.set()

    def _get_unresolved_variable(self):
        return self._unresolved.get(current_thread())

    def _set_variable_value(self, name, value):
        self._values[name] = value

    def _thread_done(self):
        """
        Removes the current thread and checks for any deadlocks.

        """

        with self._lock:
            self._threads.remove(current_thread())
            self._check_deadlock()

    def _wait_for_variable(self, name):
        """
        Waits for a variable to be defined.
        Returns immediately if it has already been defined.

        """

        with self._lock:
            if name in self._defined:
                return
            event = Event()
            self._waiters[name].append(event)
            self._check_deadlock()
        event.wait()


class MultiTemplateRenderer(object):

    def __init__(self):
        self._threads = set()
        self._var_store = VarStore(self._threads)
        self._errors = []
        self._rendered = {}

    def _render(self, source):
        try:

            # Create a variables context for the template to use.
            # Include environment variables and the var store which
            # exposes the `var.some_name` Terraform variables.
            context = os.environ.copy()
            context['var'] = self._var_store

            # Render the template.
            with open(source) as open_file:
                template = Template(
                    open_file.read(),
                    undefined=StrictUndefined,
                    keep_trailing_newline=True,
                )
            try:
                rendered = template.render(**context)
            except UndefinedError as error:
                name = self._var_store._get_unresolved_variable()
                if name:
                    error = "'var.{}' cannot be resolved".format(name)
                self._errors.append('{} in {}'.format(error, source))
                return

            # Save rendered templates.
            self._rendered[source] = rendered

            # Parse config.
            parsed = hcl.loads(rendered)

            # Process variables.
            variables = parsed.get('variable', {})
            for name, data in variables.items():
                default = data.get('default', None)
                self._var_store._define_variable(name, default)

            # Process AWS providers.
            provider = parsed.get('provider')
            if provider:
                aws = provider.get('aws')
                if aws:
                    config.aws_provider.update(aws)

            # Process S3 backents.
            terraform = parsed.get('terraform')
            if terraform:
                backend = terraform.get('backend')
                if backend:
                    s3 = backend.get('s3')
                    if s3:
                        config.s3_backend.update(s3)

        finally:
            self._var_store._thread_done()

    def add_template(self, source):
        self._threads.add(Thread(
            target=self._render,
            kwargs={'source': source},
        ))

    def set_variable_value(self, name, value):
        self._var_store._set_variable_value(name, value)

    def start(self):
        threads = self._threads.copy()
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        for error in self._errors:
            log.bad(error)
        success = not bool(self._errors)
        return (success, self._rendered)


def _populate():

    # Create a template renderer that can handle multiple files
    # with variable references between files.
    template_renderer = MultiTemplateRenderer()

    # Discover files to create in the workspace. Files in multiple
    # levels of the project directory tree with the same name will
    # be combined into a single file in the workspace.

    tfvars_files = defaultdict(set)
    tf_files = defaultdict(set)
    other_files = defaultdict(set)

    current = cwd
    while (current + '/').startswith(project_root + '/'):
        for name in sorted(os.listdir(current)):
            if name.startswith('.'):
                continue
            path = os.path.join(current, name)
            if os.path.isdir(path):
                continue
            name = name.lower()
            if name.endswith('.tfvars'):
                tfvars_files[name].add(path)
            elif name.endswith('.tf'):
                tf_files[name].add(path)
            else:
                other_files[name].add(path)
        current = os.path.dirname(current)

    # Process .tfvars files first, and read their variable values,
    # because they are required when rendering .tf files.

    for name in sorted(tfvars_files):

        source_paths = sorted(tfvars_files[name])
        target_path = os.path.join(workspace_dir, name)

        if len(source_paths) == 1:
            log.ok('copy: {}', name)
        else:
            log.ok('combine: {}', name)

        with open(target_path, 'w') as output_file:

            for source_path in source_paths:

                with open(source_path) as source_file:
                    source_file_contents = source_file.read()

                relative_source_path = os.path.relpath(source_path, project_root)
                output_file.write('# jinjaform: {}'.format(relative_source_path))
                output_file.write('\n\n')
                output_file.write(source_file_contents)
                output_file.write('\n')

                if name == 'terraform.tfvars':
                    for key, value in hcl.loads(source_file_contents).items():
                        template_renderer.set_variable_value(key, value)

    # Process .tf files as templates.

    for name in sorted(tf_files):

        source_paths = sorted(tf_files[name])

        log.ok('render: {}', name)

        for source_path in source_paths:
            template_renderer.add_template(source_path)

    success, rendered = template_renderer.start()
    if not success:
        sys.exit(1)

    for name in sorted(tf_files):

        source_paths = sorted(tf_files[name])
        target_path = os.path.join(workspace_dir, name)

        with open(target_path, 'w') as output_file:

            for source_path in source_paths:

                relative_source_path = os.path.relpath(source_path, project_root)
                output_file.write('# jinjaform: {}'.format(relative_source_path))
                output_file.write('\n\n')
                output_file.write(rendered[source_path])
                output_file.write('\n')

    # Process remaining files. Do not add source comments because
    # the file format is unknown (e.g. json files would break with #).
    for name in sorted(other_files):

        source_paths = sorted(other_files[name])
        target_path = os.path.join(workspace_dir, name)

        if len(source_paths) == 1:
            log.ok('copy: {}', name)
        else:
            log.ok('combine: {}', name)

        with open(target_path, 'w') as output_file:
            for source_path in source_paths:
                with open(source_path) as source_file:
                    output_file.write(source_file.read())


def _remove(path):
    with suppress(FileNotFoundError):
        if os.path.islink(path):
            os.remove(path)
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def clean():
    if os.path.exists(workspace_dir):
        for name in os.listdir(workspace_dir):
            if name != '.terraform':
                _remove(os.path.join(workspace_dir, name))


def create():
    # Ensure the .jinjaform/.terraform directory exists.
    os.makedirs(terraform_dir, exist_ok=True)

    # Create a shared modules directory for the entire project.
    module_cache_dir = os.path.join(jinjaform_root, 'modules')
    os.makedirs(module_cache_dir, exist_ok=True)
    module_link = os.path.join(terraform_dir, 'modules')
    _remove(module_link)
    os.symlink(module_cache_dir, module_link)

    # Create a shared plugin cache directory for the entire project.
    plugin_cache_dir = os.path.join(jinjaform_root, 'plugins')
    os.makedirs(plugin_cache_dir, exist_ok=True)
    env['TF_PLUGIN_CACHE_DIR'] = plugin_cache_dir

    # Populate workspace with Terraform configuration files.
    _populate()



#!/usr/bin/env python

from setuptools import setup

import jinjaform


setup(
    name='jinjaform',
    version=jinjaform.__version__,
    description='Jinja2 Terraform wrapper',
    author='Raymond Butcher',
    author_email='ray.butcher@claranet.uk',
    url='https://github.com/claranet/jinjaform',
    license='MIT License',
    packages=(
        'jinjaform',
    ),
    scripts=(
        'bin/jinjaform',
    ),
    install_requires=(
        'boto3',
        'botocore>=1.8.14',
        'colorama',
        'Jinja2',
        'pyhcl',
    ),
)



import os

from jinjaform import log, config


default_commands = (
    'GIT_CHECK_CLEAN',
    'GIT_CHECK_BRANCH master',
    'GIT_CHECK_REMOTE',
    'WORKSPACE_CREATE',
    'TERRAFORM_RUN',
)


def read():
    rc_path = os.path.join(config.project_root, '.jinjaformrc')
    if os.path.exists(rc_path):
        commands = []
        with open(rc_path) as open_file:
            for line in open_file:
                line = line.strip()
                if line and not line.startswith('#'):
                    commands.append(line)
    else:
        commands = default_commands

    if commands.count('WORKSPACE_CREATE') != 1:
        log.bad('configuration: WORKSPACE_CREATE must be defined once')

    if commands.count('TERRAFORM_RUN') != 1:
        log.bad('configuration: TERRAFORM_RUN must be defined once')

    if commands.index('TERRAFORM_RUN') < commands.index('WORKSPACE_CREATE'):
        log.bad('configuration: WORKSPACE_CREATE must be defined before TERRAFORM_RUN')

    for command in commands:
        parts = command.split(None, 1)
        if len(parts) == 2:
            yield parts
        else:
            yield command, None
